/*
 * Copyright 2011 by Emese Revfy <re.emese@gmail.com>
 * Copyright 2011-2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * This gcc plugin implements a defense against memory corruption attacks by preventing unwanted
 * modifications of protected memory objects.
 *
 * This plugin is only concerned with objects with static storage duration, i.e., global variables whose
 * address is fixed at runtime (dynamically allocated objects are handled by KERNSEAL).
 *
 * Homepage:
 * http://www.grsecurity.net/~ephox/const_plugin/
 *
 * Usage:
 * $ gcc -I`gcc -print-file-name=plugin`/include -fPIC -shared -O2 -o constify_plugin.so constify_plugin.c
 * $ gcc -fplugin=constify_plugin.so test.c -O2
 *
 * TODO:
 * - add no_const to constified field if they're embedded in a non-constifiable type
 */

#include "gcc-common.h"
#include "gcc-ksyms.h"

// unused C type flag in all versions 4.5-10
#define TYPE_CONSTIFY_VISITED(TYPE) TYPE_LANG_FLAG_4(TYPE)

__visible int plugin_is_GPL_compatible;

static GTY(()) tree pax_open_kernel_decl;
static bool pax_open_kernel_used_orig;
static bool pax_open_kernel_preserved_orig;
static GTY(()) tree pax_close_kernel_decl;
static bool pax_close_kernel_used_orig;
static bool pax_close_kernel_preserved_orig;

static bool enabled = true;
static bool verbose = false;

static struct plugin_info const_plugin_info = {
	.version	= "202102101420",
	.help		= "disable\tturn off constification\n",
};

#define RODATA_MUTABLE_READ_MOSTLY ".rodata..mutable.read_mostly"
#define RODATA_MUTABLE ".rodata..mutable"

static struct {
	const char *name;
	const char *asm_op;
} const_sections[] = {
	{".text",			"\t.section\t.text,\"a\""},
	{".init.rodata",		"\t.section\t.init.rodata,\"a\""},
	{".ref.rodata",			"\t.section\t.ref.rodata,\"a\""},
	{".meminit.rodata",		"\t.section\t.meminit.rodata,\"a\""},
	{".memexit.rodata",		"\t.section\t.memexit.rodata,\"a\""},
	{RODATA_MUTABLE_READ_MOSTLY,	"\t.section\t" RODATA_MUTABLE_READ_MOSTLY ",\"a\""},
	{RODATA_MUTABLE,		"\t.section\t" RODATA_MUTABLE ",\"a\""},
	{".data.once",			"\t.section\t.data.once,\"a\""},
	{".pci_fixup_early",		"\t.section\t.pci_fixup_early,\"a\""},
	{".pci_fixup_header",		"\t.section\t.pci_fixup_header,\"a\""},
	{".pci_fixup_final",		"\t.section\t.pci_fixup_final,\"a\""},
	{".pci_fixup_enable",		"\t.section\t.pci_fixup_enable,\"a\""},
	{".pci_fixup_resume",		"\t.section\t.pci_fixup_resume,\"a\""},
	{".pci_fixup_resume_early",	"\t.section\t.pci_fixup_resume_early,\"a\""},
	{".pci_fixup_suspend",		"\t.section\t.pci_fixup_suspend,\"a\""},
	{".pci_fixup_suspend_late",	"\t.section\t.pci_fixup_suspend_late,\"a\""},
	{"__param",			"\t.section\t__param,\"a\""},
};

static struct {
	const char *writable_section;
	const char *readonly_section;
} constified_sections[] = {
	{".init.data",			".init.rodata"},
	{".ref.data",			".ref.rodata"},
	{".meminit.data",		".meminit.rodata"},
	{".memexit.data",		".memexit.rodata"},
	{".data..read_mostly",		".rodata..mutable.read_mostly"},
	{".data..ro_after_init",	RODATA_MUTABLE},
	{".data..read_mostly",		RODATA_MUTABLE},
};

typedef struct {
	bool has_fptr_field;
	bool has_non_integral_field;
	bool has_writable_field;
	bool has_do_const_field;
	bool has_mutable_const_field;
	bool has_no_const_field;
	bool verbose;
} constify_info;

static const_tree get_inner_type(const_tree decl)
{
	return strip_array_types(TREE_TYPE(decl));
}

/*
 * determine whether the given structure type meets the requirements for automatic constification,
 * including the constification attributes on nested structure types
 */
static void constifiable(const_tree node, constify_info *cinfo)
{
	const_tree field;

	gcc_assert(TREE_CODE(node) == RECORD_TYPE || TREE_CODE(node) == UNION_TYPE);

	// e.g., pointer to structure fields while still constructing the structure type
	if (TYPE_FIELDS(node) == NULL_TREE)
		return;

	for (field = TYPE_FIELDS(node); field; field = TREE_CHAIN(field)) {
		const_tree type = get_inner_type(field);
		enum tree_code code = TREE_CODE(type);

		if (node == type)
			continue;

		// this is not something the frontend can produce but randomize_layout does
		if (type == void_type_node)
			continue;

		if (FUNCTION_POINTER_TYPE_P(type))
			cinfo->has_fptr_field = true;
		else if (code == RECORD_TYPE || code == UNION_TYPE) {
			if (lookup_attribute("do_const", TYPE_ATTRIBUTES(type))) {
				cinfo->has_do_const_field = true;
				if (cinfo->verbose)
					inform(DECL_SOURCE_LOCATION(field), "constify: type %qT has field %qD with 'do_const' attribute", node, field);
			} else if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type))) {
				cinfo->has_mutable_const_field = true;
				cinfo->has_writable_field = true;
			} else if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type))) {
				cinfo->has_no_const_field = true;
				cinfo->has_writable_field = true;
			} else
				constifiable(type, cinfo);
		} else {
			if (!TREE_READONLY(field))
				cinfo->has_writable_field = true;
			if (!INTEGRAL_TYPE_P(type))
				cinfo->has_non_integral_field = true;
		}
	}
}

static bool check_attr_usage(tree type, tree name, const char *incompat_attr1, const char *incompat_attr2)
{
	if (!TYPE_P(type)) {
		error("%qE attribute applies to types only (%qD)", name, type);
		return false;
	}

	if (TREE_CODE(type) != RECORD_TYPE && TREE_CODE(type) != UNION_TYPE && TREE_CODE(type) != ARRAY_TYPE) {
		error("%qE attribute used on %qT applies to struct and union types only", name, type);
		return false;
	}

	if (lookup_attribute(IDENTIFIER_POINTER(name), TYPE_ATTRIBUTES(type))) {
		error("%qE attribute used on %qT is already applied to the type", name, type);
		return false;
	}

	if (lookup_attribute(incompat_attr1, TYPE_ATTRIBUTES(type))) {
		error("%qE attribute used on %qT is incompatible with '%s'", name, type, incompat_attr1);
		return false;
	}

	if (incompat_attr2 && lookup_attribute(incompat_attr2, TYPE_ATTRIBUTES(type))) {
		error("%qE attribute used on %qT is incompatible with '%s'", name, type, incompat_attr2);
		return false;
	}

	return true;
}

/*
 * the no_const attribute on a type ensures that the type won't be automatically constified.
 */
static tree handle_no_const_attribute(tree *node, tree name, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	tree *type;

	*no_add_attrs = true;
	if (TREE_CODE(*node) == FUNCTION_DECL) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to functions (%qF)", name, *node);
		return NULL_TREE;
	}

	if (TREE_CODE(*node) == PARM_DECL) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to function parameters (%qD)", name, *node);
		return NULL_TREE;
	}

	if (VAR_P(*node)) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to variables (%qD)", name, *node);
		return NULL_TREE;
	}

	if (TYPE_P(*node)) {
		type = node;
		gcc_assert(flags & (int)ATTR_FLAG_TYPE_IN_PLACE);
	} else {
		if (TREE_CODE(*node) != TYPE_DECL) {
			error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to %qD (%qT)", name, *node, TREE_TYPE(*node));
			return NULL_TREE;
		}
		gcc_assert(!(flags & (int)ATTR_FLAG_TYPE_IN_PLACE));
		type = &TREE_TYPE(*node);
	}

	if (TYPE_P(*node)) {
		if (check_attr_usage(*type, name, "do_const", "mutable_const"))
			*no_add_attrs = false;
		return NULL_TREE;
	}

	if (!check_attr_usage(*type, name, "do_const", NULL))
		return NULL_TREE;

	if (!TYPE_P(*node)) {
		*type = build_variant_type_copy(*type);
		TYPE_ATTRIBUTES(*type) = tree_cons(name, NULL_TREE, TYPE_ATTRIBUTES(*type));
	}

	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(*type))) {
		*no_add_attrs = false;
		return NULL_TREE;
	}

	if (enabled && TYPE_FIELDS(*type))
		error("1 %qE attribute used on type %qT that is not constified", name, *type);

	return NULL_TREE;
}

static void constify_type(tree type)
{
	tree t;

	gcc_assert(type == TYPE_MAIN_VARIANT(type));
	gcc_assert(type == TYPE_CANONICAL(type));
	TYPE_READONLY(type) = 1;
	C_TYPE_FIELDS_READONLY(type) = 1;
	TYPE_CONSTIFY_VISITED(type) = 1;

	gcc_assert(type == TYPE_CANONICAL(type));

	// fix up any already existing equivalent constified types
	for (t = TYPE_MAIN_VARIANT(type); t; t = TYPE_NEXT_VARIANT(t)) {
		tree pointer_to;

		if (t == type)
			continue;

		if (TYPE_QUALS(t) != TYPE_QUALS(type))
			continue;

		TYPE_CANONICAL(t) = TYPE_CANONICAL(type);

		if (!TYPE_POINTER_TO(t)) {
			gcc_assert(!TYPE_POINTER_TO(type));
			continue;
		}

		if (TYPE_POINTER_TO(type))
			continue;

		for (pointer_to = TYPE_POINTER_TO(t); pointer_to; pointer_to = TYPE_NEXT_PTR_TO(pointer_to))
			TREE_TYPE(pointer_to) = type;

		TYPE_POINTER_TO(type) = TYPE_POINTER_TO(t);
	}
}

/*
 * the do_const attribute on a type ensures that the type will be forcibly constified (as if
 * the type declaration had been replaced with a const typedef).
 */
static tree handle_do_const_attribute(tree *node, tree name, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	*no_add_attrs = true;

	if (!check_attr_usage(*node, name, "no_const", "mutable_const"))
		return NULL_TREE;

	*no_add_attrs = false;
	return NULL_TREE;
}

/*
 * the mutable_const attribute on a type ensures that all objects of the type with static storage
 * duration will be forcibly constified.
 */
static tree handle_mutable_const_attribute(tree *node, tree name, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	*no_add_attrs = true;

	if (TYPE_P(*node) && !check_attr_usage(*node, name, "no_const", "do_const"))
		return NULL_TREE;
	else if (DECL_P(*node) && !VAR_P(*node)) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute applies to variables only, not %qD", name, *node);
		return NULL_TREE;
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

static struct attribute_spec no_const_attr = ATTRIBUTE_SPEC_INIT("no_const", 0, 0, false, false, false, handle_no_const_attribute, true, NULL);
static struct attribute_spec do_const_attr = ATTRIBUTE_SPEC_INIT("do_const", 0, 0, false, true, false, handle_do_const_attribute, true, NULL);
static struct attribute_spec mutable_const_attr = ATTRIBUTE_SPEC_INIT("mutable_const", 0, 0, false, false, false, handle_mutable_const_attribute, true, NULL);

static void register_attributes(void *event_data __unused, void *data __unused)
{
	register_attribute(&no_const_attr);
	register_attribute(&do_const_attr);
	register_attribute(&mutable_const_attr);
}

static void finish_type(void *event_data, void *data __unused)
{
	tree type = (tree)event_data;
	constify_info cinfo = {
		.has_fptr_field = false,
		.has_non_integral_field = false,
		.has_writable_field = false,
		.has_do_const_field = false,
		.has_mutable_const_field = false,
		.has_no_const_field = false,
		.verbose = false,
	};

	if (type == NULL_TREE || type == error_mark_node)
		return;

#if BUILDING_GCC_VERSION >= 5000
	if (TREE_CODE(type) == ENUMERAL_TYPE)
		return;
#endif

	if (TYPE_FIELDS(type) == NULL_TREE || TYPE_CONSTIFY_VISITED(type))
		return;

	constifiable(type, &cinfo);

	// handle mutable_const: don't allow mixing with do_const fields
	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type))) {
		if (cinfo.has_do_const_field)
			error("'mutable_const' attribute used on type %qT that is partially constified", type);
		else {
			TYPE_CONSTIFY_VISITED(type) = 1;
		}
		return;
	}

	// handle no_const: don't allow on types that wouldn't be constified anyway
	if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type))) {
		if ((cinfo.has_fptr_field && !cinfo.has_writable_field) || cinfo.has_do_const_field || cinfo.has_mutable_const_field) {
			TYPE_CONSTIFY_VISITED(type) = 1;
		} else
			error("'no_const' attribute used on type %qT that is not constified", type);
		return;
	}

	// handle do_const: don't allow on types that are eligible for automatic constification, even if they lack fptr fields
	if (lookup_attribute("do_const", TYPE_ATTRIBUTES(type))) {
		if (!cinfo.has_writable_field) {
			error("'do_const' attribute used on type %qT that is%sautomatically constified", type, cinfo.has_fptr_field ? " " : " not ");
			return;
		}
		constify_type(type);
		return;
	}

	// propagate mutable_const from fields
	if (!cinfo.has_non_integral_field && !cinfo.has_no_const_field && !cinfo.has_do_const_field && cinfo.has_mutable_const_field) {
		TYPE_CONSTIFY_VISITED(type) = 1;
//		add_type_attr(type, "mutable_const", NULL_TREE);
		return;
	}

	// automatic constification of types with fptrs and immutable fields
	if ((cinfo.has_fptr_field || cinfo.has_do_const_field) && !cinfo.has_writable_field) {
		constify_type(type);
		add_type_attr(type, "do_const", NULL_TREE);
		return;
	}

	if (cinfo.has_do_const_field) {
		error("non-constifiable type %qT has field with 'do_const' attribute", type);
		cinfo.has_fptr_field = false,
		cinfo.has_non_integral_field = false,
		cinfo.has_writable_field = false,
		cinfo.has_do_const_field = false,
		cinfo.has_mutable_const_field = false,
		cinfo.has_no_const_field = false,
		cinfo.verbose = true;
		constifiable(type, &cinfo);
	}
	TYPE_CONSTIFY_VISITED(type) = 1;
}

static bool is_const_section(const char *name)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(const_sections); i++)
		if (!strcmp(const_sections[i].name, name))
			return true;

	return false;
}

static bool is_constified_var(varpool_node_ptr node)
{
	const_tree var = NODE_DECL(node);
	const_tree type = get_inner_type(var);
	const char *name;

	if (node->alias)
		return false;

	if (TREE_READONLY(var))
		return false;

	if (lookup_attribute("mutable_const", DECL_ATTRIBUTES(var)))
		return true;

	switch (TREE_CODE(type)) {
	default:
		return false;

	case RECORD_TYPE:
	case UNION_TYPE:
		if (DECL_EXTERNAL(var))
			return false;
		break;
	}

	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type)))
		return true;

	name = get_decl_section_name(var);
	if (name && is_const_section(name))
		return false;

	if (!TYPE_READONLY(type) || !C_TYPE_FIELDS_READONLY(type))
		return false;

	return TYPE_CONSTIFY_VISITED(type);
}

static void handle_section_mismatch(varpool_node_ptr node)
{
	tree var = NODE_DECL(node);
	location_t loc = DECL_SOURCE_LOCATION(var);
	const_tree type = get_inner_type(var);
	const_tree section;
	const char *name;
	size_t i;

	name = get_decl_section_name(var);
	section = lookup_attribute("section", DECL_ATTRIBUTES(var));

	// check that the two ways of storing section info are consistently set/unset
	if (!section != !name) {
		fprintf(stderr, "DECL_SECTION: [%s] != section attr: [%s]\n", name ? name : "", section ? TREE_STRING_POINTER(section) : "");
		dump_varpool_node(stderr, node);
		debug_tree(var);
		gcc_unreachable();
	}

	// the implicit default section is .data
	if (!name) {
		if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type)))
			return;
		if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type))) {
			//if (lookup_attribute("mutable_const", DECL_ATTRIBUTES(var)))
			//	error_at(loc, "constify: variable %qD with constified mutable type %qT has mutable_const attribute", var, type);
			if (verbose)
				inform(loc, "constify: variable %qD with constified mutable type %qT moved into section " RODATA_MUTABLE, var, type);
			set_decl_section_name(var, RODATA_MUTABLE);
		} else if (lookup_attribute("mutable_const", DECL_ATTRIBUTES(var))) {
			if (verbose)
				inform(loc, "constify: constified mutable variable %qD moved into section " RODATA_MUTABLE, var);
			set_decl_section_name(var, RODATA_MUTABLE);
		} else if (POINTER_TYPE_P(type) && lookup_attribute("do_const", TYPE_ATTRIBUTES(TREE_TYPE(type))) != NULL_TREE) {
			if (verbose)
				inform(loc, "constify: pointer to constified variable %qD moved into section " RODATA_MUTABLE, var);
			set_decl_section_name(var, RODATA_MUTABLE);
		} else if (POINTER_TYPE_P(type) && lookup_attribute("mutable_const", TYPE_ATTRIBUTES(TREE_TYPE(type))) != NULL_TREE) {
			if (verbose)
				inform(loc, "constify: pointer to mutable variable %qD moved into section " RODATA_MUTABLE, var);
			set_decl_section_name(var, RODATA_MUTABLE);
		} else if (!TYPE_CONSTIFY_VISITED(type)) {
			dump_varpool_node(stderr, node);
			debug_tree(var);
			debug_tree(type);
			gcc_unreachable();
		}
		return;
	}

//	gcc_assert(!TREE_CHAIN(section));
	gcc_assert(TREE_VALUE(section));

	section = TREE_VALUE(TREE_VALUE(section));
	gcc_assert(!strcmp(TREE_STRING_POINTER(section), name));

	if (is_const_section(name))
		return;

	for (i = 0; i < ARRAY_SIZE(constified_sections); i++) {
		if (strcmp(constified_sections[i].writable_section, name))
			continue;

		if (verbose)
			inform(loc, "constify: constified mutable variable %qD moved into section %s", var, constified_sections[i].readonly_section);
		set_decl_section_name(var, constified_sections[i].readonly_section);
		return;
	}

	dump_varpool_node(stderr, node);
	debug_tree(var);
	debug_tree(var);
	fflush(stderr);
	error_at(loc, "constified variable %qD placed into writable section %s", var, name);
}

/*
 * this works around a gcc bug/feature where uninitialized globals
 * are moved into the .bss section regardless of any constification
 * see gcc/varasm.c:bss_initializer_p()
 */
static void fix_initializer(tree var)
{
	tree type = TREE_TYPE(var);

	if (DECL_INITIAL(var))
		return;

	DECL_INITIAL(var) = build_constructor(type, NULL);
}

static void handle_global_variables(void *event_data __unused, void *data __unused)
{
	varpool_node_ptr node;

	FOR_EACH_VARIABLE(node) {
		if (!is_constified_var(node))
			continue;

		handle_section_mismatch(node);
		fix_initializer(NODE_DECL(node));
	}
}

/*
 * the TU must provide the open/close wrappers. look them up by name early enough and mark them as used
 * otherwise gcc would delete them if the TU itself doesn't use them and before the instrumentation needs them
 */
static void find_open_close_functions(void *event_data __unused, void *data __unused)
{
	tree fndecl = (tree)event_data;
	const char *asmname;

	if (fndecl == error_mark_node)
		return;

	if (TREE_CODE(fndecl) != FUNCTION_DECL)
		return;

	if (DECL_ARTIFICIAL(fndecl))
		return;

	if (DECL_ABSTRACT_ORIGIN(fndecl) && DECL_ABSTRACT_ORIGIN(fndecl) != fndecl)
		return;

	asmname = DECL_NAME_POINTER(fndecl);
	gcc_assert(asmname[0]);

	if (DECL_NAME_LENGTH(fndecl) == sizeof("pax_open_kernel") - 1 &&
	    id_equal(DECL_NAME(fndecl), "pax_open_kernel")) {
		gcc_assert(!pax_open_kernel_decl);
		pax_open_kernel_decl = fndecl;
		pax_open_kernel_used_orig = TREE_USED(fndecl);
		TREE_USED(fndecl) = 1;
		pax_open_kernel_preserved_orig = DECL_PRESERVE_P(fndecl);
		DECL_PRESERVE_P(fndecl) = 1;
		return;
	}

	if (DECL_NAME_LENGTH(fndecl) == sizeof("pax_close_kernel") - 1 &&
	    id_equal(DECL_NAME(fndecl), "pax_close_kernel")) {
		gcc_assert(!pax_close_kernel_decl);
		pax_close_kernel_decl = fndecl;
		pax_close_kernel_used_orig = TREE_USED(fndecl);
		TREE_USED(fndecl) = 1;
		pax_close_kernel_preserved_orig = DECL_PRESERVE_P(fndecl);
		DECL_PRESERVE_P(fndecl) = 1;
		return;
	}
}

/*
 * undo the forced used marks on the open/close wrappers
 * the instrumentation no longer needs them as we're past inlining here
 */
static void unmark_open_close_functions(void *event_data __unused, void *data __unused)
{
	cgraph_node_ptr node;

	if (!pax_open_kernel_decl) {
		gcc_assert(!pax_close_kernel_decl);
		return;
	}
	gcc_assert(pax_close_kernel_decl);

#if BUILDING_GCC_VERSION <= 4007
	if (!gimple_has_body_p(pax_open_kernel_decl))
#endif
	{
		node = cgraph_get_node(pax_open_kernel_decl);
		gcc_assert(node);
#if BUILDING_GCC_VERSION <= 4007
		gcc_assert(NODE_SYMBOL(node)->needed);
		NODE_SYMBOL(node)->needed = pax_open_kernel_preserved_orig;
#else
		gcc_assert(NODE_SYMBOL(node)->force_output);
		NODE_SYMBOL(node)->force_output = pax_open_kernel_preserved_orig;
#endif
		TREE_USED(pax_open_kernel_decl) = pax_open_kernel_used_orig;
		DECL_PRESERVE_P(pax_open_kernel_decl) = pax_open_kernel_preserved_orig;
	}

#if BUILDING_GCC_VERSION <= 4007
	if (!gimple_has_body_p(pax_close_kernel_decl))
#endif
	{
		node = cgraph_get_node(pax_close_kernel_decl);
		gcc_assert(node);
		TREE_USED(pax_close_kernel_decl) = pax_close_kernel_used_orig;
		DECL_PRESERVE_P(pax_close_kernel_decl) = pax_close_kernel_preserved_orig;
#if BUILDING_GCC_VERSION <= 4007
		gcc_assert(NODE_SYMBOL(node)->needed);
		NODE_SYMBOL(node)->needed = pax_close_kernel_preserved_orig;
#else
		gcc_assert(NODE_SYMBOL(node)->force_output);
		NODE_SYMBOL(node)->force_output = pax_close_kernel_preserved_orig;
#endif
	}
}

static unsigned int check_local_variables_execute(void)
{
	unsigned int ret = 0;
	tree var;

	unsigned int i;

	FOR_EACH_LOCAL_DECL(cfun, i, var) {
		tree type = TREE_TYPE(var);

		gcc_assert(DECL_P(var));
		if (is_global_var(var))
			continue;

		if (TREE_CODE(type) != RECORD_TYPE && TREE_CODE(type) != UNION_TYPE)
			continue;

		if (!TYPE_READONLY(type) || !C_TYPE_FIELDS_READONLY(type))
			continue;

		if (!TYPE_CONSTIFY_VISITED(type))
			continue;

		error_at(DECL_SOURCE_LOCATION(var), "constified variable %qE cannot be local", var);
		ret = 1;
	}
	return ret;
}

#define PASS_NAME check_local_variables
#define NO_GATE
#include "gcc-generate-gimple-pass.h"

enum states {
	invalid,
	seen_n,
	seen_nm,
	seen_nmo,
	seen_nmoc,
	seen_nmocm,
	seen_no,
	seen_noc,
	seen_nocm,
	num_states
};

enum events {
	not_write,
	mutable_write,
	unsafe_write,
	open_kernel,
	close_kernel,
	num_events
};

struct bb_state {
	gimple_stmt_ptr ins_open;
	gimple_stmt_ptr ins_close;
	gimple_stmt_ptr src_close;
	bool has_unsafe_write;
	enum states state;
};

static bool tainted[0];

static void gimple_set_not_write(gimple_stmt_ptr stmt)
{
	gimple_set_plf(stmt, GF_PLF_1, false);
	gimple_set_plf(stmt, GF_PLF_2, false);
}

static bool gimple_set_not_write_p(gimple_stmt_ptr stmt)
{
	return !gimple_plf(stmt, GF_PLF_1) && !gimple_plf(stmt, GF_PLF_2);
}

static void gimple_set_mutable_write(gimple_stmt_ptr stmt)
{
	gimple_set_plf(stmt, GF_PLF_1, false);
	gimple_set_plf(stmt, GF_PLF_2, true);
}

static bool gimple_set_mutable_write_p(gimple_stmt_ptr stmt)
{
	return !gimple_plf(stmt, GF_PLF_1) && gimple_plf(stmt, GF_PLF_2);
}

static void gimple_set_unsafe_write(gimple_stmt_ptr stmt)
{
	gimple_set_plf(stmt, GF_PLF_1, true);
	gimple_set_plf(stmt, GF_PLF_2, false);
}

static bool gimple_set_unsafe_write_p(gimple_stmt_ptr stmt)
{
	return gimple_plf(stmt, GF_PLF_1) && !gimple_plf(stmt, GF_PLF_2);
}

static void gimple_set_event(gimple_stmt_ptr stmt, enum events event)
{
	switch (event) {
	case open_kernel:
	case close_kernel:
	case not_write:
		gimple_set_not_write(stmt);
		return;

	case mutable_write:
		gimple_set_mutable_write(stmt);
		return;

	case unsafe_write:
		gimple_set_unsafe_write(stmt);
		return;

	case num_events:
		gcc_unreachable();
	}
}

static bool is_pax_close_kernel(gimple_stmt_ptr stmt);
static bool is_pax_open_kernel(gimple_stmt_ptr stmt);

static enum events gimple_get_event(gimple_stmt_ptr stmt)
{
	if (gimple_set_not_write_p(stmt)) {
		if (is_pax_open_kernel(stmt))
			return open_kernel;
		if (is_pax_close_kernel(stmt))
			return close_kernel;
		return not_write;
	}

	if (gimple_set_mutable_write_p(stmt))
		return mutable_write;

	if (gimple_set_unsafe_write_p(stmt))
		return unsafe_write;

	gcc_unreachable();
}

static bool is_write(gimple_stmt_ptr stmt)
{
	if (gimple_clobber_p(stmt))
		return false;

	return gimple_store_p(stmt);
}

/*
 * determine if type is mutable
 */
enum mutable_type { mutable_yes, mutable_no, mutable_unknown };

static enum mutable_type is_mutable_type(const_tree type)
{
	if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type)))
		return mutable_no;

	if (lookup_attribute("do_const", TYPE_ATTRIBUTES(type)))
		return mutable_yes;

	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type)))
		return mutable_yes;

	return mutable_unknown;
}

/*
 * determine if var is a mutable object
 */
static bool is_mutable_var(const_tree var)
{
	const_tree vartype = TREE_TYPE(var);
	const char *name;

	gcc_assert(VAR_P(var));

	if (!is_global_var(var))
		return false;

	if (is_mutable_type(vartype) == mutable_yes)
		return true;

	if (lookup_attribute("mutable_const", DECL_ATTRIBUTES(var)))
		return true;

	name = get_decl_section_name(var);
	return name && is_const_section(name);
}

/*
 * determine if arg can be a reference to a mutable object
 */
static bool is_mutable_arg(const_tree arg)
{
	switch (TREE_CODE(arg)) {
	default:
		debug_tree(arg);
		gcc_unreachable();

	case ADDR_EXPR:
	case COMPONENT_REF: {
		tree argtype;

		argtype = TREE_TYPE(arg);
		if (!POINTER_TYPE_P(argtype))
			return false;
		if (is_mutable_type(TREE_TYPE(argtype)) == mutable_yes)
			return true;

		arg = get_inner_arg(TREE_OPERAND(arg, 0));
#if BUILDING_GCC_VERSION == 4005
		if (TREE_CODE(arg) == INDIRECT_REF)
#else
		if (TREE_CODE(arg) == MEM_REF)
#endif
			return is_mutable_type(TREE_TYPE(arg)) == mutable_yes;

		if (TREE_CODE(arg) == VAR_DECL)
			return is_mutable_var(arg);
		return false;
	}

	case INTEGER_CST: {
		tree argtype;

		argtype = TREE_TYPE(arg);
		if (!POINTER_TYPE_P(argtype))
			return false;
		return is_mutable_type(TREE_TYPE(argtype)) == mutable_yes;
	}

	case SSA_NAME: {
		gimple_stmt_ptr def_stmt;
		tree argtype, rhs1;

		argtype = TREE_TYPE(arg);
		if (!POINTER_TYPE_P(argtype))
			return false;
		if (is_mutable_type(TREE_TYPE(argtype)) == mutable_yes)
			return true;

		def_stmt = SSA_NAME_DEF_STMT(arg);

		if (!is_gimple_assign(def_stmt))
			return false;

		rhs1 = gimple_assign_rhs1(def_stmt);
		switch (gimple_assign_rhs_code(def_stmt)) {
		default:
			fprintf(stderr, "ISMUTABLEARG ASSIGN %s ", get_tree_code_name(gimple_assign_rhs_code(def_stmt)));
			debug_gimple_stmt(def_stmt);
			debug_tree(arg);
			debug_tree(SSA_NAME_VAR(arg));
			debug_tree(rhs1);
			gcc_unreachable();

		case ADDR_EXPR:
			argtype = TREE_TYPE(argtype);
			if (POINTER_TYPE_P(argtype))
				if (is_mutable_type(TREE_TYPE(argtype)) == mutable_yes)
					return true;

			return is_mutable_arg(rhs1);

		case ARRAY_REF:
		case COMPONENT_REF:
		case INDIRECT_REF:
#if BUILDING_GCC_VERSION >= 4006
		case MEM_REF:
#endif
		case PARM_DECL:
			argtype = TREE_TYPE(argtype);
			if (POINTER_TYPE_P(argtype))
				return is_mutable_type(TREE_TYPE(argtype)) == mutable_yes;

			return false;

		case INTEGER_CST:
			return false;

		case CONVERT_EXPR:
		case NOP_EXPR:
			if (!POINTER_TYPE_P(TREE_TYPE(rhs1)))
				return false;
			return is_mutable_arg(rhs1);

		case POINTER_PLUS_EXPR:
		case SSA_NAME:
			return is_mutable_arg(rhs1);

		case VAR_DECL:
			return is_mutable_var(rhs1);
		}
	}

	case VAR_DECL:
		return is_mutable_var(arg);
	}
}

static enum events is_mutable_lhs(gimple_stmt_ptr stmt, tree lhs)
{
	tree ref, offset, reftype;
	poly_int64_pod bitsize, bitpos;
	enum machine_mode mode;
	int unsignedp, reversep, volatilep;

	ref = get_inner_reference(lhs, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
	reftype = TREE_TYPE(ref);

	switch (TREE_CODE(ref)) {
	default:
		print_gimple_stmt(stderr, stmt, 0, TDF_LINENO);
		debug_tree(lhs);
		debug_tree(ref);
		gcc_unreachable();

	case VAR_DECL:
		if (is_mutable_var(ref))
			return mutable_write;
		return INTEGRAL_TYPE_P(TREE_TYPE(lhs)) ? not_write : unsafe_write;

	case INDIRECT_REF:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
		switch (is_mutable_type(reftype)) {
		default:
			gcc_unreachable();

		case mutable_yes:
			return mutable_write;

		case mutable_no:
			return unsafe_write;

		case mutable_unknown:
			return is_mutable_arg(TREE_OPERAND(ref, 0)) ? mutable_write : unsafe_write;
		}

	// simple parameter write
	case PARM_DECL:
		if (!offset)
			return not_write;
		gcc_unreachable();

	// simple return value write
	case RESULT_DECL:
	case SSA_NAME:
		return not_write;
	}
}

enum placement {before, after};

static gimple_stmt_ptr constify_insert_call(tree callee, gimple_stmt_iterator *gsi, gimple_stmt_ptr reference, enum placement placement)
{
	gimple_stmt_ptr stmt;
	gcall *gcallee;
	cgraph_node_ptr node;
	int frequency;
	basic_block bb;

	// insert call to callee
	stmt = gimple_build_call(callee, 0);
	gimple_set_location(stmt, gimple_location(reference));
	gimple_set_block(stmt, gimple_block(reference));
	gcallee = as_a_gcall(stmt);
	if (placement == before)
		gsi_insert_before(gsi, gcallee, GSI_SAME_STMT);
	else
		gsi_insert_after(gsi, gcallee, GSI_CONTINUE_LINKING);

	// update the cgraph
	bb = gimple_bb(gcallee);
	node = cgraph_get_node(callee);
	gcc_assert(node);
	frequency = compute_call_stmt_bb_frequency(current_function_decl, bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), node, gcallee, bb->count, frequency, bb->loop_depth);

	return stmt;
}

static bool is_pax_close_kernel(gimple_stmt_ptr stmt)
{
	if (pax_close_kernel_decl == NULL_TREE)
		return false;

	if (!is_gimple_call(stmt))
		return false;

	return gimple_call_fndecl(stmt) == pax_close_kernel_decl;
}

static bool is_pax_open_kernel(gimple_stmt_ptr stmt)
{
	if (pax_open_kernel_decl == NULL_TREE)
		return false;

	if (!is_gimple_call(stmt))
		return false;

	return gimple_call_fndecl(stmt) == pax_open_kernel_decl;
}

static void pax_open_kernel(gimple_stmt_iterator *gsi, gimple_stmt_ptr reference, enum placement placement)
{
	gimple_stmt_ptr stmt;

	if (!pax_open_kernel_decl) {
		error_at(gimple_location(reference), "pax_open_kernel must be declared");
		return;
	}

	if (verbose) {
		location_t loc = gsi_end_p(*gsi) ? gimple_location(reference) : gimple_location(gsi_stmt(*gsi));
		inform(loc, "constify: writable access to constified region begins %s", placement == before ? "before" : "after");
	}

	stmt = constify_insert_call(pax_open_kernel_decl, gsi, reference, placement);
	gimple_set_event(stmt, open_kernel);
}

static void pax_close_kernel(gimple_stmt_iterator *gsi, gimple_stmt_ptr reference, enum placement placement)
{
	gimple_stmt_ptr stmt;

	if (!pax_close_kernel_decl) {
		error_at(gimple_location(reference), "pax_close_kernel must be declared");
		return;
	}

	if (verbose) {
		location_t loc = gsi_end_p(*gsi) ? gimple_location(reference) : gimple_location(gsi_stmt(*gsi));
		inform(loc, "constify: writable access to constified region ends %s%s", placement == before ? "before" : "after", gsi_end_p(*gsi) ? " (approximately)" : "");
	}

	stmt = constify_insert_call(pax_close_kernel_decl, gsi, reference, placement);
	gimple_set_event(stmt, close_kernel);
}

/*
 * determine if stmt can modify a mutable object and thus needs an open kernel to do so
 *
 * a mutable object is either placed into a read-only section or has its type marked with
 * the mutable_const attribute so we have to find pointer dereferences of such types.
 */
static enum events constify_classify_stmt(gimple_stmt_ptr stmt)
{
	tree ref, offset, reftype;
	poly_int64_pod bitsize, bitpos;
	enum machine_mode mode;
	int unsignedp, reversep, volatilep;

	if (pax_open_kernel_decl && is_pax_open_kernel(stmt))
		return open_kernel;

	if (pax_close_kernel_decl && is_pax_close_kernel(stmt))
		return close_kernel;

	switch (gimple_code(stmt)) {
	default:
		error_at(gimple_location(stmt), "constify: CLASSIFY %s", gimple_code_name[gimple_code(stmt)]);
		debug_gimple_stmt(stmt);
		gcc_unreachable();

	case GIMPLE_ASM: {
		gasm *asm_stmt;
		const char *asm_string;
		size_t i, noutputs;

		asm_stmt = as_a_gasm(stmt);
		asm_string = gimple_asm_string(asm_stmt);
		if (!*asm_string)
			return not_write;

		if (strstr(asm_string, "rap_safe_asm"))
			return not_write;

		return unsafe_write;

		noutputs = gimple_asm_noutputs(asm_stmt);
		for (i = 0; i < noutputs; ++i) {
			tree op, arg;

			op = gimple_asm_output_op(asm_stmt, i);
			arg = TREE_VALUE(op);
			ref = get_inner_reference(arg, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
			reftype = TREE_TYPE(ref);

			switch (TREE_CODE(ref)) {
			default:
				debug_gimple_stmt(stmt);
				debug_tree(op);
				debug_tree(arg);
				debug_tree(ref);
				gcc_unreachable();

			case SSA_NAME:
				if (!POINTER_TYPE_P(reftype))
					continue;
				if (is_mutable_type(TREE_TYPE(reftype)) == mutable_yes)
					return mutable_write;
				continue;

			case COMPONENT_REF:
#if BUILDING_GCC_VERSION >= 4006
			case MEM_REF:
#endif
				if (is_mutable_arg(TREE_OPERAND(ref, 0)))
					return mutable_write;
				continue;

			case VAR_DECL:
				if (is_mutable_var(ref))
					return mutable_write;
				continue;
			}
		}

		return unsafe_write;
	}

	case GIMPLE_ASSIGN:
		return is_write(stmt) ? is_mutable_lhs(stmt, gimple_get_lhs(stmt)) : not_write;

	case GIMPLE_CALL: {
		tree lhs, fndecl, count, param;
		const char *fnname;
		unsigned int i, e, flags;

		lhs = gimple_call_lhs(stmt);
		if (lhs && is_mutable_lhs(stmt, lhs) == mutable_write) {
			tree var;
			gimple_stmt_ptr assign;
			gimple_stmt_iterator gsi;

			var = create_tmp_var(TREE_TYPE(lhs), "constify");
			add_referenced_var(var);
			mark_sym_for_renaming(var);

			assign = gimple_build_assign(lhs, var);
			gimple_set_location(assign, gimple_location(stmt));
			gimple_set_block(assign, gimple_block(stmt));
			gsi = gsi_for_stmt(stmt);
			gsi_insert_after(&gsi, assign, GSI_SAME_STMT);

			gimple_call_set_lhs(stmt, var);
			unlink_stmt_vdef(stmt);
			update_stmt(stmt);
			lhs = var;
		}

		fndecl = gimple_call_fndecl(stmt);

		if (fndecl == NULL_TREE)
			return unsafe_write;

		flags = gimple_call_flags(stmt);

		if (flags & ECF_CONST || flags & ECF_PURE)
			return not_write;

		fnname = DECL_NAME_POINTER(fndecl);
		if (!strcmp(fnname, "memcpy") ||
		    !strcmp(fnname, "memset") ||
		    gimple_call_builtin_p(stmt, BUILT_IN_MEMCPY) ||
		    gimple_call_builtin_p(stmt, BUILT_IN_MEMSET))
			count = gimple_call_arg(stmt, 2);
		else if (!strcmp(fnname, "__memzero"))
			count = gimple_call_arg(stmt, 1);
		else
			return unsafe_write;

		if (integer_zerop(count))
			return not_write;
		else if (TREE_CODE(count) == SSA_NAME) {
			gimple_stmt_ptr def_stmt = SSA_NAME_DEF_STMT(count);
			if (gimple_assign_single_p(def_stmt) && integer_zerop(gimple_assign_rhs1(def_stmt)))
				return not_write;
		}

		return is_mutable_arg(gimple_call_arg(stmt, 0)) ? mutable_write : unsafe_write;

		param = TYPE_ARG_TYPES(gimple_call_fntype(stmt));
		if (param == NULL_TREE)
			return unsafe_write;

		for (i = 0, e = gimple_call_num_args(stmt); i < e; ++i, param = param ? TREE_CHAIN(param) : NULL_TREE) {
			tree arg;

			arg = gimple_call_arg(stmt, i);
			gcc_assert(!param || POINTER_TYPE_P(TREE_TYPE(arg)) == POINTER_TYPE_P(TREE_VALUE(param)));
			if (!POINTER_TYPE_P(TREE_TYPE(arg)))
				continue;
			if (param && is_mutable_type(TREE_TYPE(TREE_VALUE(param))) == mutable_yes)
				return unsafe_write;
			if (is_mutable_type(TREE_TYPE(TREE_TYPE(arg))) == mutable_yes)
				return unsafe_write;
			if (CONSTANT_CLASS_P(arg))
				continue;
			if (is_mutable_arg(arg))
				return unsafe_write;
			continue;
		}

		return not_write;
	}

	case GIMPLE_COND:
	case GIMPLE_DEBUG:
	case GIMPLE_GOTO:
	case GIMPLE_LABEL:
	case GIMPLE_NOP:
	case GIMPLE_PREDICT:
	case GIMPLE_RETURN:
	case GIMPLE_SWITCH:
		return not_write;
	}

	gcc_unreachable();
}

static gimple_stmt_ptr constify_remove_stmt(gimple_stmt_ptr stmt)
{
	gimple_stmt_iterator gsi;

	location_t loc;
	tree block;

	gsi = gsi_for_stmt(stmt);
	loc = gimple_location(stmt);
	block = gimple_block(stmt);

	unlink_stmt_vdef(stmt);

	stmt = gimple_build_nop();
	gimple_set_location(stmt, loc);
	gimple_set_block(stmt, block);
	gsi_replace(&gsi, stmt, true);
	gimple_set_event(stmt, not_write);

	return stmt;
}

/*
 * the workhorse: find and instrument mutable writes that would fail otherwise at runtime
 *
 * the instrumentation surrounds the mutable writes with pax_open_kernel/pax_close_kernel
 * calls. there's some light effort at optimizing the number and placement of these calls
 * as their performance impact is unfortunately not negligible with current processor features.
 */
static enum events constify_step(struct bb_state *bb_state, gimple_stmt_ptr stmt, enum events event)
{
	switch (event) {
	case not_write:
		return not_write;

	case mutable_write:
		switch (bb_state->state) {
		case invalid:
			gcc_unreachable();

		case seen_n:
			bb_state->state = seen_nm;

			bb_state->ins_open = stmt;
			bb_state->ins_close = stmt;
			return mutable_write;

		case seen_no:
		case seen_nmo:
			return mutable_write;

		case seen_nm:
		case seen_nocm:
		case seen_nmocm:
			bb_state->ins_close = stmt;
			return mutable_write;

		case seen_noc:
			bb_state->state = seen_nocm;

			gcc_assert(bb_state->src_close);
			gcc_assert(stmt != bb_state->src_close);
			gcc_assert(bb_state->ins_close == NULL);

			constify_remove_stmt(bb_state->src_close);
			bb_state->ins_close = stmt;
			bb_state->src_close = NULL;
			return mutable_write;

		case seen_nmoc:
			bb_state->state = seen_nmocm;

			gcc_assert(bb_state->src_close);
			gcc_assert(stmt != bb_state->src_close);

			constify_remove_stmt(bb_state->src_close);
			bb_state->ins_close = stmt;
			bb_state->src_close = NULL;
			return mutable_write;

		case num_states:
			gcc_unreachable();
		}
		gcc_unreachable();

	case unsafe_write:
		switch (bb_state->state) {
		case invalid:
			gcc_unreachable();

		case seen_n:
			return unsafe_write;

		case seen_nm: {
			gimple_stmt_iterator gsi2;

			bb_state->state = seen_n;

			gcc_assert(bb_state->ins_open);
			gsi2 = gsi_for_stmt(bb_state->ins_open);
			pax_open_kernel(&gsi2, stmt, before);
			bb_state->ins_open = NULL;

			gcc_assert(bb_state->ins_close);
			gsi2 = gsi_for_stmt(bb_state->ins_close);
			pax_close_kernel(&gsi2, stmt, after);
			bb_state->ins_close = NULL;
			return unsafe_write;
		}

		case seen_no:
		case seen_nmo:
			return mutable_write;

		case seen_noc:
		case seen_nmoc:
			bb_state->state = seen_n;

			gcc_assert(bb_state->src_close);
			bb_state->src_close = NULL;
			gcc_assert(bb_state->ins_close == NULL);
			return unsafe_write;

		case seen_nocm:
		case seen_nmocm: {
			gimple_stmt_iterator gsi2;

			bb_state->state = seen_n;

			gcc_assert(bb_state->ins_close);
			gsi2 = gsi_for_stmt(bb_state->ins_close);
			pax_close_kernel(&gsi2, stmt, after);
			bb_state->ins_close = NULL;
			return unsafe_write;
		}

		case num_states:
			gcc_unreachable();
		}
		gcc_unreachable();

	case open_kernel:
		switch (bb_state->state) {
		case invalid:
			gcc_unreachable();

		case seen_n:
			bb_state->state = seen_no;
			return open_kernel;

		case seen_nm:
			bb_state->state = seen_nmo;

			constify_remove_stmt(stmt);
			bb_state->ins_close = NULL;
			return open_kernel;

		case seen_no:
		case seen_nmo:
			error_at(gimple_location(stmt), "constify: pax_open_kernel encountered after a previous pax_open_kernel");
			gcc_unreachable();

		case seen_noc:
			bb_state->state = seen_no;

			constify_remove_stmt(stmt);

			gcc_assert(bb_state->src_close);
			constify_remove_stmt(bb_state->src_close);
			bb_state->src_close = NULL;

			gcc_assert(bb_state->ins_close == NULL);
			return open_kernel;

		case seen_nocm:
			bb_state->state = seen_no;

			constify_remove_stmt(stmt);
			gcc_assert(bb_state->ins_close);
			bb_state->ins_close = NULL;
			return open_kernel;

		case seen_nmoc:
			bb_state->state = seen_nmo;

			constify_remove_stmt(stmt);
			gcc_assert(!bb_state->ins_close);
			return open_kernel;

		case seen_nmocm:
			bb_state->state = seen_nmo;

			constify_remove_stmt(stmt);
			gcc_assert(bb_state->ins_close);
			bb_state->ins_close = NULL;
			return open_kernel;

		case num_states:
			gcc_unreachable();
		}
		gcc_unreachable();

	case close_kernel:
		switch (bb_state->state) {
		case invalid:
			gcc_unreachable();

		case seen_n:
		case seen_nm:
			error_at(gimple_location(stmt), "constify: pax_close_kernel encountered without a previous pax_open_kernel");
			gcc_unreachable();

		case seen_noc:
		case seen_nmoc:
		case seen_nocm:
		case seen_nmocm:
			error_at(gimple_location(stmt), "constify: pax_close_kernel encountered after a previous pax_close_kernel");
			gcc_unreachable();

		case seen_no:
			bb_state->state = seen_noc;

			gcc_assert(!bb_state->src_close);
			bb_state->src_close = stmt;
			return close_kernel;

		case seen_nmo:
			bb_state->state = seen_nmoc;

			gcc_assert(!bb_state->src_close);
			bb_state->src_close = stmt;
			gcc_assert(!bb_state->ins_close);
			return close_kernel;

		case num_states:
			gcc_unreachable();
		}
		gcc_unreachable();

	case num_events:
		gcc_unreachable();
	}
	gcc_unreachable();
}

/*
 * the main analysis run over all BBs in reverse postorder to determine where to place
 * potential open/close instrumentation or emit some of them directly when there's no
 * choice for placement.
 */
static struct bb_state *analyze_mutable_writes(int *postorder, unsigned int postorder_num)
{
	unsigned int i;
	struct bb_state *bb_states;

	bb_states = XNEWVEC(struct bb_state, last_basic_block_for_fn(cfun));
	for (i = last_basic_block_for_fn(cfun); i; i--) {
		bb_states[i - 1].ins_open = NULL;
		bb_states[i - 1].ins_close = NULL;
		bb_states[i - 1].src_close = NULL;
		bb_states[i - 1].has_unsafe_write = false;
		bb_states[i - 1].state = invalid;
	}

	for (i = postorder_num; i; i--) {
		unsigned int j;
		basic_block bb;
		edge e;
		edge_iterator ei;
		gimple_stmt_ptr stmt;
		gimple_stmt_iterator gsi;
		struct bb_state *bb_state;
		sbitmap pred_states;

		j = postorder[i - 1];
		bb = BASIC_BLOCK_FOR_FN(cfun, j);
		bb_state = &bb_states[j];

		pred_states = sbitmap_alloc(num_states);
		sbitmap_clear(pred_states);

		FOR_EACH_EDGE(e, ei, bb->preds) {
			enum states state;

			state = bb_states[e->src->index].state;
			if (state == invalid)
				continue;

			sbitmap_set_bit(pred_states, state);
		}

		switch (sbitmap_count_bits(pred_states)) {
		case 0:
			gcc_assert(j == ENTRY_BLOCK || j == EXIT_BLOCK);
			bb_state->state = seen_n;
			break;

		case 1: {
			enum states state = (enum states)sbitmap_first_set_bit(pred_states);

			switch (state) {
			default:
				debug_bb(bb);
				debug_sbitmap(pred_states);
				gcc_unreachable();

			case seen_n:
			case seen_no:
			case seen_nmo:
				bb_state->state = state;
				break;

			case seen_nmocm:
				gcc_assert(j == EXIT_BLOCK);
				bb_state->state = state;
				break;

			case seen_nm:
			case seen_nocm:
				FOR_EACH_EDGE(e, ei, bb->preds) {
					if (bb_states[e->src->index].state == invalid)
						continue;

					gcc_assert(bb_states[e->src->index].ins_close);
					constify_step(&bb_states[e->src->index], bb_states[e->src->index].ins_close, unsafe_write);
				}
				bb_state->state = seen_n;
				break;

			case seen_noc:
			case seen_nmoc:
				FOR_EACH_EDGE(e, ei, bb->preds) {
					if (bb_states[e->src->index].state == invalid)
						continue;

					gcc_assert(!bb_states[e->src->index].ins_close);
					constify_step(&bb_states[e->src->index], NULL, unsafe_write);
				}
				bb_state->state = seen_n;
				break;
			}
			break;
		}

		default:
			FOR_EACH_EDGE(e, ei, bb->preds) {
				enum states state;

				state = bb_states[e->src->index].state;
				switch (state) {
				default:
					debug_sbitmap(pred_states);
					gcc_unreachable();

				case seen_n:
					continue;

				case seen_nm:
					gcc_assert(bb_states[e->src->index].ins_close);
					constify_step(&bb_states[e->src->index], bb_states[e->src->index].ins_close, unsafe_write);
					continue;

				case seen_noc:
				case seen_nmoc:
					gcc_assert(!bb_states[e->src->index].ins_close);
					constify_step(&bb_states[e->src->index], NULL, unsafe_write);
					continue;
				}
			}

			bb_state->state = seen_n;
			break;
		}

		sbitmap_free(pred_states);

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			enum events event;

			stmt = gsi_stmt(gsi);
			event = constify_classify_stmt(stmt);
			gimple_set_event(stmt, constify_step(bb_state, stmt, event));

			if (event != unsafe_write)
				continue;

			bb_state->has_unsafe_write = true;

			if (!bb->loop_father)
				continue;

			bb->loop_father->aux = tainted;
		}
	}

	return bb_states;
}

static basic_block loop_single_exit_block(loop_p loop)
{
	struct loop_exit *exit;
	basic_block bb_single_exit = NULL;

	exit = loop->exits->next;
	if (exit->e) {
		bb_single_exit = exit->e->dest;

		for (exit = exit->next; exit->e; exit = exit->next)
			if (bb_single_exit != exit->e->dest)
				return NULL;
	}

	return bb_single_exit;
}

static bool instrument_mutable_writes(int *postorder, unsigned int postorder_num, struct bb_state *bb_states)
{
	unsigned int i;
	loop_p loop;
	bool changed = false;
#if BUILDING_GCC_VERSION <= 4008
	loop_iterator li;
#endif

	for (i = postorder_num; i; i--) {
		unsigned int j;
		gimple_stmt_iterator gsi;
		struct bb_state *bb_state;

		j = postorder[i - 1];
		bb_state = &bb_states[j];

		if (bb_state->ins_open) {
			gsi = gsi_for_stmt(bb_state->ins_open);
			pax_open_kernel(&gsi, bb_state->ins_open, before);
			bb_state->ins_open = NULL;
			changed = true;
		}

		if (bb_state->ins_close) {
			gsi = gsi_for_stmt(bb_state->ins_close);
			pax_close_kernel(&gsi, bb_state->ins_close, after);
			bb_state->ins_close = NULL;
			changed = true;
		}
	}

#if BUILDING_GCC_VERSION <= 4008
	FOR_EACH_LOOP(li, loop, LI_FROM_INNERMOST) {
#else
	FOR_EACH_LOOP(loop, LI_FROM_INNERMOST) {
#endif
		basic_block *bbs;
		unsigned int opens, closes;
		gimple_stmt_iterator gsi;
		basic_block bb_open, bb_close;
		gimple_stmt_ptr stmt;

		if (loop->aux == tainted)
			continue;

		opens = closes = 0;
		bbs = get_loop_body(loop);

		for (i = 0; i < loop->num_nodes; i++) {
			for (gsi = gsi_start_bb(bbs[i]); !gsi_end_p(gsi); gsi_next(&gsi)) {
				stmt = gsi_stmt(gsi);
				switch (gimple_get_event(stmt)) {
				default:
					continue;

				case unsafe_write:
					gcc_unreachable();

				case open_kernel:
					constify_remove_stmt(stmt);
					opens++;
					continue;

				case close_kernel:
					constify_remove_stmt(stmt);
					closes++;
					continue;
				}
			}
		}

		free(bbs);

		gcc_assert(opens == closes);
		if (opens == 0 && closes == 0)
			continue;

		bb_open = get_immediate_dominator(CDI_DOMINATORS, loop->header);
		gcc_assert(single_succ_p(bb_open));
		gsi = gsi_last_nondebug_bb(bb_open);
		if (gsi_end_p(gsi)) {
			gcc_assert(single_succ_p(bb_open));
			stmt = gsi_stmt(gsi_last_nondebug_bb(single_succ(bb_open)));
		} else
			stmt = gsi_stmt(gsi);
		gcc_assert(stmt);
		pax_open_kernel(&gsi, stmt, gsi_end_p(gsi) ? before : after);

		bb_close = loop_single_exit_block(loop);
		if (bb_close && dominated_by_p(CDI_DOMINATORS, bb_close, bb_open)) {
			gcc_assert(dominated_by_p(CDI_POST_DOMINATORS, bb_open, bb_close));
			gsi = gsi_start_nondebug_after_labels_bb(bb_close);
			gcc_assert(!gsi_end_p(gsi));
			pax_close_kernel(&gsi, gsi_stmt(gsi), before);
		} else {
#if BUILDING_GCC_VERSION <= 4007
			VEC(edge, heap) *exits;
#else
			VEC(edge, heap) exits;
#endif
			unsigned int j;
			edge e;

			exits = get_loop_exit_edges(loop);
#if BUILDING_GCC_VERSION <= 4007
			FOR_EACH_VEC_ELT(edge, exits, j, e) {
#else
			FOR_EACH_VEC_ELT(exits, j, e) {
#endif
				basic_block new_close;

				gcc_assert(e->src->loop_father == loop);

				gsi = gsi_last_nondebug_bb(e->dest);
				gcc_assert(!gsi_end_p(gsi));
				stmt = gsi_stmt(gsi);

				new_close = plugin_split_edge(e);
				gcc_assert(single_pred_p(new_close));
				gsi = gsi_start_nondebug_after_labels_bb(new_close);
				pax_close_kernel(&gsi, stmt, after);
			}
#if BUILDING_GCC_VERSION <= 4007
			VEC_free(edge, heap, exits);
#else
			exits.release();
#endif
		}

		changed = true;
	}

	return changed;
}

static void propagate_loop_taint(int *postorder, unsigned int postorder_num)
{
	loop_p loop;
#if BUILDING_GCC_VERSION <= 4008
	loop_iterator li;
#endif

#if BUILDING_GCC_VERSION <= 4008
	FOR_EACH_LOOP(li, loop, LI_FROM_INNERMOST) {
#else
	FOR_EACH_LOOP(loop, LI_FROM_INNERMOST) {
#endif
		loop_p outer;

		if (loop->aux != tainted)
			continue;

		for (outer = loop_outer(loop); outer; outer = loop_outer(outer))
			outer->aux = tainted;
	}
}

static unsigned int instrument_mutable_writes_execute(void)
{
	unsigned int round;
	bool changed;
	loop_p loop;
#if BUILDING_GCC_VERSION <= 4008
	loop_iterator li;
#endif

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

#if BUILDING_GCC_VERSION <= 4008
	FOR_EACH_LOOP(li, loop, LI_INCLUDE_ROOT)
		loop->aux = NULL;
#else
	FOR_EACH_LOOP(loop, LI_INCLUDE_ROOT)
		loop->aux = NULL;
#endif

	/*
	 * the core loop that iterates over analysis and instrumentation.
	 */
	changed = true;
	for (round = 0; changed; round++) {
		int *postorder;
		unsigned int postorder_num;
		struct bb_state *bb_states;

		if (verbose && dump_file)
			fprintf(dump_file, "\n;; start round %u for %s\n", round, DECL_NAME_POINTER(current_function_decl));

		postorder = XNEWVEC(int, n_basic_blocks_for_fn(cfun));
		postorder_num = post_order_compute(postorder, true, false);
		gcc_assert(ENTRY_BLOCK == postorder[postorder_num - 1]);

		bb_states = analyze_mutable_writes(postorder, postorder_num);

		if (round == 0)
			propagate_loop_taint(postorder, postorder_num);

		changed = instrument_mutable_writes(postorder, postorder_num, bb_states);
		XDELETEVEC(bb_states);

		XDELETEVEC(postorder);

		if (verbose && dump_file) {
			dump_flow_info(dump_file, dump_flags);
			fprintf(dump_file, "\n;; end round %u for %s\n\n", round, DECL_NAME_POINTER(current_function_decl));
		}
	}

#if BUILDING_GCC_VERSION <= 4008
	FOR_EACH_LOOP(li, loop, LI_INCLUDE_ROOT)
#else
	FOR_EACH_LOOP(loop, LI_INCLUDE_ROOT)
#endif
		loop->aux = NULL;

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

static bool instrument_mutable_writes_gate(void)
{
	tree section;

	if (MODULE != NULL_TREE)
		return true;

	section = lookup_attribute("section", DECL_ATTRIBUTES(current_function_decl));
	if (!section || !TREE_VALUE(section))
		return true;

	section = TREE_VALUE(TREE_VALUE(section));

	gcc_assert(strncmp(TREE_STRING_POINTER(section), ".vsyscall_", 10));

	if (!strncmp(TREE_STRING_POINTER(section), ".init.text", 10))
		return false;
	if (!strncmp(TREE_STRING_POINTER(section), ".cpuinit.text", 13))
		return false;
	if (!strncmp(TREE_STRING_POINTER(section), ".meminit.text", 13))
		return false;
	if (!strncmp(TREE_STRING_POINTER(section), ".head.text", 10))
		return false;

	return true;
}

#define PASS_NAME instrument_mutable_writes
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_ggc_collect | TODO_rebuild_cgraph_edges | TODO_verify_flow
#include "gcc-generate-gimple-pass.h"

static unsigned int (*old_section_type_flags)(tree decl, const char *name, int reloc);

static unsigned int constify_section_type_flags(tree decl, const char *name, int reloc)
{
	return is_const_section(name) ? 0 : old_section_type_flags(decl, name, reloc);
}

static bool (*old_lang_hooks_init)(void);

static bool constify_lang_hooks_init(void)
{
	register_callback("constify_plugin", PLUGIN_FINISH_TYPE, finish_type, NULL);
	return old_lang_hooks_init();
}

static tree (*old_merge_decl_attributes)(tree, tree);

static void constify_check_merged_decl_attributes(tree olddecl, tree newdecl)
{
	const char *oldname, *newname;
	bool oldattr, newattr;

	if (!VAR_P(olddecl) || !VAR_P(newdecl))
		return;

	oldname = get_decl_section_name(olddecl);
	newname = get_decl_section_name(newdecl);
	oldattr = !!lookup_attribute("mutable_const", DECL_ATTRIBUTES(olddecl));
	newattr = !!lookup_attribute("mutable_const", DECL_ATTRIBUTES(newdecl));

	if (oldattr == newattr) {
		if (!oldname || !newname)
			return;

		if (!strcmp(oldname, newname))
			return;
	}

	error_at(DECL_SOURCE_LOCATION(newdecl), "conflicting declarations for %qD [%s]", newdecl, newname ?: "<default>");
	inform(DECL_SOURCE_LOCATION(olddecl), "previous declaration of %q+D was here [%s]", olddecl, oldname ?: "<default>");
}

static tree constify_merge_decl_attributes(tree olddecl, tree newdecl)
{
	constify_check_merged_decl_attributes(olddecl, newdecl);
	return old_merge_decl_attributes(olddecl, newdecl);
}

static void constify_start_unit(void *gcc_data __unused, void *user_data __unused)
{
	old_section_type_flags = targetm.section_type_flags;
	targetm.section_type_flags = constify_section_type_flags;

	old_merge_decl_attributes = targetm.merge_decl_attributes;
	targetm.merge_decl_attributes = constify_merge_decl_attributes;
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	int i;

	static const struct ggc_root_tab gt_ggc_r_gt_constify[] = {
		{
			.base = &pax_open_kernel_decl,
			.nelt = 1,
			.stride = sizeof(pax_open_kernel_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		{
			.base = &pax_close_kernel_decl,
			.nelt = 1,
			.stride = sizeof(pax_close_kernel_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		LAST_GGC_ROOT_TAB
	};

	PASS_INFO(check_local_variables, "ssa", 1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(instrument_mutable_writes, "ssa", 1, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!(strcmp(argv[i].key, "disable"))) {
			enabled = false;
			continue;
		}
		if (!(strcmp(argv[i].key, "verbose"))) {
			verbose = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	if (strncmp(lang_hooks.name, "GNU C", 5) || !strncmp(lang_hooks.name, "GNU C+", 6)) {
		inform(UNKNOWN_LOCATION, G_("%s supports C only, not %s"), plugin_name, lang_hooks.name);
		enabled = false;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &const_plugin_info);
	if (enabled) {
		old_lang_hooks_init = lang_hooks.init;
		lang_hooks.init = constify_lang_hooks_init;
		register_callback(plugin_name, PLUGIN_START_UNIT, constify_start_unit, (void *)plugin_name);
		register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, find_open_close_functions, NULL);
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_START, handle_global_variables, NULL);
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_END, unmark_open_close_functions, NULL);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &check_local_variables_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &instrument_mutable_writes_pass_info);
		register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_constify);
	}
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	return 0;
}
